/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "math.h"
#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>

#include <fstream>
#include <windows.h> //for WideCharToMultiByte


using namespace std;

#define PN_INPUT1 0
#define PN_INPUT2 1
#define PN_INPUT3 2
#define PN_INPUT4 3
#define PN_INPUT5 4

#define PN_FILENAME 5
#define PN_BUFFER 6
#define PN_INTERPOLATION 7
#define PN_IFFTINTERPOLATION 8
#define PN_MODE 9

#define PN_OUTPUT1 10
#define PN_OUTPUT2 11

const powtwo[17] = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536};

static const double pi = 3.1415926535897932384626433832795;
static const double pi2 = 6.283185307179586476925286766559;
static const double pih = 1.5707963267948966192313216916398;
float denormal = 1.0e-18;


void Module::sub_process(long buffer_offset, long sampleFrames )
{
    float *in1  = buffer_offset + input1_buffer;
    float *in2  = buffer_offset + input2_buffer;
    float *in3  = buffer_offset + input3_buffer;
    float *in4  = buffer_offset + input4_buffer;
    float *in5  = buffer_offset + input5_buffer;

    float *out1 = buffer_offset + output1_buffer;
    float *out2 = buffer_offset + output2_buffer;

    while(--sampleFrames >= 0) {
		if (*in1 > 0.f && gatebuf <= 0.f) {
			if (left && good) {
				bpf = p0 = 0;
				p0b = n;
				p1b = 0;
				play = 1;
				first = 1;
				for (i = 0; i < n; i++) phl0[i] = phr0[i] = phl1[i] = phr1[i] = 0.f;
				pf = *in4 * samples;
				p = (int) pf;
				p /= n;	p *= n;
				pf = (float) p;
			}
			else play = 0;
			for (i = hb; i < n; i++) ol1[i] = or1[i] = 0.f;
		}

		if (*in2 != pitchbuf) {
			inc = pow(2.f, *in2 * 10.f) * incrate;
			pitchbuf = *in2;
		}

		if (*in3 != timebuf) {
			tinc = pow(2.f, *in3 * 10.f) * incrate;
			timebuf = *in3;
		}

		if (play > 0) {
			p0 < hb ? p1 = p0 + hb : p1 = p0 - hb;

			if (p0 < p0b) {
				p = (int) pf;
				pfr = p / hb;
				unsigned long int fa = pfr * n;
				o = *in5;
				while (o < 0.f) o += 1.f;
				while (o >= 1.f) o -= 1.f;
				switch (ifftinterpolation) {
					case 0:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								srl = rl[sh0] + shd + (rl[sh1] - rl[sh0]);
								srr = rr[sh0] + shd + (rr[sh1] - rr[sh0]);
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = srl * cos(phl0[i]);
								ilb[i] = srl * -sin(phl0[i]);
								rrb[i] = srr * cos(phr0[i]);
								irb[i] = srr * -sin(phr0[i]);
							}
						}
						else {
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = rl[j] * cos(phl0[i]);
								ilb[i] = rl[j] * -sin(phl0[i]);
								rrb[i] = rr[j] * cos(phr0[i]);
								irb[i] = rr[j] * -sin(phr0[i]);
							}
						}
						break;
					case 1:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								o0 = rl[sh0] + o * (rl[sh0 + n] - rl[sh0]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[sh0] + o * (rr[sh0 + n] - rr[sh0]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								o4 = rl[sh1] + o * (rl[sh1 + n] - rl[sh1]);
								o5 = rr[sh1] + o * (rr[sh1 + n] - rr[sh1]);
								o0 += shd * (o4 - o0);
								o2 += shd * (o5 - o2);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl0[i]);
								ilb[i] = o0 * -sin(phl0[i]);
								rrb[i] = o2 * cos(phr0[i]);
								irb[i] = o2 * -sin(phr0[i]);
							}
						}
						else {
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								o0 = rl[j] + o * (rl[j + n] - rl[j]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[j] + o * (rr[j + n] - rr[j]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl0[i]);
								ilb[i] = o0 * -sin(phl0[i]);
								rrb[i] = o2 * cos(phr0[i]);
								irb[i] = o2 * -sin(phr0[i]);
							}
						}
						break;
				}
				dofft(rlb, ilb);
				dofft(rrb, irb);

				if (first) {	//	no envelope if first frame
					for (i = 0; i < hb; i++) {
						ol0[i] = rlb[i] * invn;
						or0[i] = rrb[i] * invn;
					}
					for (i = hb; i < n; i++) {
						ol0[i] = rlb[i] * invn * sine[i];
						or0[i] = rrb[i] * invn * sine[i];
					}
					first = 0;
				}
				else {
					for (i = 0; i < n; i++) {
						ol0[i] = rlb[i] * invn * sine[i];
						or0[i] = rrb[i] * invn * sine[i];
					}
				}
			}
			else if (p1 < p1b) {
				p = (int) pf;
				pfr = p / hb;
				unsigned long int fa = pfr * n;
				o = *in5;
				while (o < 0.f) o += 1.f;
				while (o >= 1.f) o -= 1.f;
				switch (ifftinterpolation) {
					case 0:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								srl = rl[sh0] + shd + (rl[sh1] - rl[sh0]);
								srr = rr[sh0] + shd + (rr[sh1] - rr[sh0]);
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = srl * cos(phl1[i]);
								ilb[i] = srl * -sin(phl1[i]);
								rrb[i] = srr * cos(phr1[i]);
								irb[i] = srr * -sin(phr1[i]);
							}
						}
						else {
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = rl[j] * cos(phl1[i]);
								ilb[i] = rl[j] * -sin(phl1[i]);
								rrb[i] = rr[j] * cos(phr1[i]);
								irb[i] = rr[j] * -sin(phr1[i]);
							}
						}
						break;
					case 1:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								o0 = rl[sh0] + o * (rl[sh0 + n] - rl[sh0]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[sh0] + o * (rr[sh0 + n] - rr[sh0]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								o4 = rl[sh1] + o * (rl[sh1 + n] - rl[sh1]);
								o5 = rr[sh1] + o * (rr[sh1 + n] - rr[sh1]);
								o0 += shd * (o4 - o0);
								o2 += shd * (o5 - o2);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl1[i]);
								ilb[i] = o0 * -sin(phl1[i]);
								rrb[i] = o2 * cos(phr1[i]);
								irb[i] = o2 * -sin(phr1[i]);
							}
						}
						else {
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								o0 = rl[j] + o * (rl[j + n] - rl[j]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[j] + o * (rr[j + n] - rr[j]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl1[i]);
								ilb[i] = o0 * -sin(phl1[i]);
								rrb[i] = o2 * cos(phr1[i]);
								irb[i] = o2 * -sin(phr1[i]);
							}
						}
						break;
				}
				dofft(rlb, ilb);
				dofft(rrb, irb);

				for (i = 0; i < n; i++) {
					ol1[i] = rlb[i] * invn * sine[i];
					or1[i] = rrb[i] * invn * sine[i];
				}
			}
			p0b = p0;
			p1b = p1;

			switch (interpolation) {
				case 0:
					*out1 = ol0[p0] + ol1[p1];
					*out2 = or0[p0] + or1[p1];
					break;
				case 1:
					bd = bpf - p0;
					p0 < nm1 ? s1 = p0 + 1 : s1 = p0;
					p1 < nm1 ? z1 = p1 + 1 : z1 = p1;
					*out1 = ol0[p0] + ol1[p1] + bd * ((ol0[s1] - ol0[p0]) + (ol1[z1] - ol1[p1]));
					*out2 = or0[p0] + or1[p1] + bd * ((or0[s1] - or0[p0]) + (or1[z1] - or1[p1]));
					break;
				case 2:
					bd = bpf - p0;
					p0 > 0 ? s0 = p0 - 1 : s0 = 0;
					s1 = p0;
					s1 < nm1 ? s2 = s1 + 1 : s2 = s1;
					s2 < nm1 ? s3 = s2 + 1 : s3 = s2;

					p1 > 0 ? z0 = p1 - 1 : z0 = 0;
					z1 = p1;
					z1 < nm1 ? z2 = z1 + 1 : z2 = z1;
					z2 < nm1 ? z3 = z2 + 1 : z3 = z2;

					f0 = (ol0[s3] - ol0[s2]) - (ol0[s0] - ol0[s1]);
					f1 = (ol0[s0] - ol0[s1]) - f0;
					f2 = ol0[s2] - ol0[s0];
					o0 = ((f0 * bd + f1) * bd + f2) * bd + ol0[s1];

					f0 = (ol1[z3] - ol1[z2]) - (ol1[z0] - ol1[z1]);
					f1 = (ol1[z0] - ol1[z1]) - f0;
					f2 = ol1[z2] - ol1[z0];
					o1 = ((f0 * bd + f1) * bd + f2) * bd + ol1[z1];

					f0 = (or0[s3] - or0[s2]) - (or0[s0] - or0[s1]);
					f1 = (or0[s0] - or0[s1]) - f0;
					f2 = or0[s2] - or0[s0];
					o2 = ((f0 * bd + f1) * bd + f2) * bd + or0[s1];

					f0 = (or1[z3] - or1[z2]) - (or1[z0] - or1[z1]);
					f1 = (or1[z0] - or1[z1]) - f0;
					f2 = or1[z2] - or1[z0];
					o3 = ((f0 * bd + f1) * bd + f2) * bd + or1[z1];

					*out1 = o0 + o1;
					*out2 = o2 + o3;
					break;
			}

			bpf += inc;
			if (bpf > n) bpf -= n;
			p0 = (int) bpf;

			pf += tinc;
			if (pf >= samples) play = 0;
		}
		else wake = *out1 = *out2 = 0.f;
		gatebuf = *in1;

		in1++;	in2++;	in3++;	in4++;	in5++;
		out1++;	out2++;
    }
	if (!wake && !wakebuf) {
		getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_STATIC);
		getPin(PN_OUTPUT2)->TransmitStatusChange( SampleClock(), ST_STATIC);
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module::sub_process_static);
	}
    wakebuf = wake;
}









void Module::sub_process_scan(long buffer_offset, long sampleFrames )
{
    float *in1  = buffer_offset + input1_buffer;
    float *in2  = buffer_offset + input2_buffer;
    float *in3  = buffer_offset + input3_buffer;
    float *in4  = buffer_offset + input4_buffer;
    float *in5  = buffer_offset + input5_buffer;

    float *out1 = buffer_offset + output1_buffer;
    float *out2 = buffer_offset + output2_buffer;

    while(--sampleFrames >= 0) {
		if (*in1 > 0.f && gatebuf <= 0.f) {
			if (left && good) {
				if (!play) {
					bpf = p0 = 0;
					p0b = n;
					p1b = 0;
					play = 1;
					first = 1;
					for (i = 0; i < n; i++) phl0[i] = phr0[i] = phl1[i] = phr1[i] = 0.f;
				}
			}
			else play = 0;
			for (i = hb; i < n; i++) ol1[i] = or1[i] = 0.f;
		}

		if (*in2 != pitchbuf) {
			inc = pow(2.f, *in2 * 10.f) * incrate;
			pitchbuf = *in2;
		}

		if (*in3 != timebuf) {
			tinc = pow(2.f, *in3 * 10.f) * incrate;
			timebuf = *in3;
		}

		if (play > 0) {
			p0 < hb ? p1 = p0 + hb : p1 = p0 - hb;

			if (p0 < p0b) {
				pf = *in4 * samples;
				p = (int) pf;
				pfr = p / hb;
				unsigned long int fa = pfr * n;
				o = *in5;
				while (o < 0.f) o += 1.f;
				while (o >= 1.f) o -= 1.f;
				switch (ifftinterpolation) {
					case 0:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								srl = rl[sh0] + shd + (rl[sh1] - rl[sh0]);
								srr = rr[sh0] + shd + (rr[sh1] - rr[sh0]);
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = srl * cos(phl0[i]);
								ilb[i] = srl * -sin(phl0[i]);
								rrb[i] = srr * cos(phr0[i]);
								irb[i] = srr * -sin(phr0[i]);
							}
						}
						else {
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = rl[j] * cos(phl0[i]);
								ilb[i] = rl[j] * -sin(phl0[i]);
								rrb[i] = rr[j] * cos(phr0[i]);
								irb[i] = rr[j] * -sin(phr0[i]);
							}
						}
						break;
					case 1:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								o0 = rl[sh0] + o * (rl[sh0 + n] - rl[sh0]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[sh0] + o * (rr[sh0 + n] - rr[sh0]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								o4 = rl[sh1] + o * (rl[sh1 + n] - rl[sh1]);
								o5 = rr[sh1] + o * (rr[sh1 + n] - rr[sh1]);
								o0 += shd * (o4 - o0);
								o2 += shd * (o5 - o2);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl0[i]);
								ilb[i] = o0 * -sin(phl0[i]);
								rrb[i] = o2 * cos(phr0[i]);
								irb[i] = o2 * -sin(phr0[i]);
							}
						}
						else {
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								o0 = rl[j] + o * (rl[j + n] - rl[j]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[j] + o * (rr[j + n] - rr[j]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl0[i]);
								ilb[i] = o0 * -sin(phl0[i]);
								rrb[i] = o2 * cos(phr0[i]);
								irb[i] = o2 * -sin(phr0[i]);
							}
						}
						break;
				}
				dofft(rlb, ilb);
				dofft(rrb, irb);

				if (first) {	//	no envelope if first frame
					for (i = 0; i < hb; i++) {
						ol0[i] = rlb[i] * invn;
						or0[i] = rrb[i] * invn;
					}
					for (i = hb; i < n; i++) {
						ol0[i] = rlb[i] * invn * sine[i];
						or0[i] = rrb[i] * invn * sine[i];
					}
					first = 0;
				}
				else {
					for (i = 0; i < n; i++) {
						ol0[i] = rlb[i] * invn * sine[i];
						or0[i] = rrb[i] * invn * sine[i];
					}
				}
			}
			else if (p1 < p1b) {
				pf = *in4 * samples;
				p = (int) pf;
				pfr = p / hb;
				unsigned long int fa = pfr * n;
				o = *in5;
				while (o < 0.f) o += 1.f;
				while (o >= 1.f) o -= 1.f;
				switch (ifftinterpolation) {
					case 0:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								srl = rl[sh0] + shd + (rl[sh1] - rl[sh0]);
								srr = rr[sh0] + shd + (rr[sh1] - rr[sh0]);
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = srl * cos(phl1[i]);
								ilb[i] = srl * -sin(phl1[i]);
								rrb[i] = srr * cos(phr1[i]);
								irb[i] = srr * -sin(phr1[i]);
							}
						}
						else {
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								phl[i] += il[j];
								phr[i] += ir[j];
								rlb[i] = rl[j] * cos(phl1[i]);
								ilb[i] = rl[j] * -sin(phl1[i]);
								rrb[i] = rr[j] * cos(phr1[i]);
								irb[i] = rr[j] * -sin(phr1[i]);
							}
						}
						break;
					case 1:
						if (o > 0.f) {
							o *= n;	shift = (int) o;
							shd = o - (float) shift;
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								sh0 = i + shift;	sh0 %= n;
								sh0 == nm1 ? sh1 = 0 : sh1 = sh0 + 1;
								sh0 += fa;	sh1 += fa;
								o0 = rl[sh0] + o * (rl[sh0 + n] - rl[sh0]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[sh0] + o * (rr[sh0 + n] - rr[sh0]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								o4 = rl[sh1] + o * (rl[sh1 + n] - rl[sh1]);
								o5 = rr[sh1] + o * (rr[sh1 + n] - rr[sh1]);
								o0 += shd * (o4 - o0);
								o2 += shd * (o5 - o2);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl1[i]);
								ilb[i] = o0 * -sin(phl1[i]);
								rrb[i] = o2 * cos(phr1[i]);
								irb[i] = o2 * -sin(phr1[i]);
							}
						}
						else {
							pfrd = p % hb;
							o = (float) pfrd / (float) hb;
							for (i = 0; i < n; i++) {	//	polar to cartesian
								j = fa + i;
								o0 = rl[j] + o * (rl[j + n] - rl[j]);
								o1 = il[j] + o * (il[j + n] - il[j]);
								o2 = rr[j] + o * (rr[j + n] - rr[j]);
								o3 = ir[j] + o * (ir[j + n] - ir[j]);
								phl[i] += o1;
								phr[i] += o3;
								rlb[i] = o0 * cos(phl1[i]);
								ilb[i] = o0 * -sin(phl1[i]);
								rrb[i] = o2 * cos(phr1[i]);
								irb[i] = o2 * -sin(phr1[i]);
							}
						}
						break;
				}
				dofft(rlb, ilb);
				dofft(rrb, irb);

				for (i = 0; i < n; i++) {
					ol1[i] = rlb[i] * invn * sine[i];
					or1[i] = rrb[i] * invn * sine[i];
				}
			}
			p0b = p0;
			p1b = p1;

			switch (interpolation) {
				case 0:
					*out1 = ol0[p0] + ol1[p1];
					*out2 = or0[p0] + or1[p1];
					break;
				case 1:
					bd = bpf - p0;
					p0 < nm1 ? s1 = p0 + 1 : s1 = p0;
					p1 < nm1 ? z1 = p1 + 1 : z1 = p1;
					*out1 = ol0[p0] + ol1[p1] + bd * ((ol0[s1] - ol0[p0]) + (ol1[z1] - ol1[p1]));
					*out2 = or0[p0] + or1[p1] + bd * ((or0[s1] - or0[p0]) + (or1[z1] - or1[p1]));
					break;
				case 2:
					bd = bpf - p0;
					p0 > 0 ? s0 = p0 - 1 : s0 = 0;
					s1 = p0;
					s1 < nm1 ? s2 = s1 + 1 : s2 = s1;
					s2 < nm1 ? s3 = s2 + 1 : s3 = s2;

					p1 > 0 ? z0 = p1 - 1 : z0 = 0;
					z1 = p1;
					z1 < nm1 ? z2 = z1 + 1 : z2 = z1;
					z2 < nm1 ? z3 = z2 + 1 : z3 = z2;

					f0 = (ol0[s3] - ol0[s2]) - (ol0[s0] - ol0[s1]);
					f1 = (ol0[s0] - ol0[s1]) - f0;
					f2 = ol0[s2] - ol0[s0];
					o0 = ((f0 * bd + f1) * bd + f2) * bd + ol0[s1];

					f0 = (ol1[z3] - ol1[z2]) - (ol1[z0] - ol1[z1]);
					f1 = (ol1[z0] - ol1[z1]) - f0;
					f2 = ol1[z2] - ol1[z0];
					o1 = ((f0 * bd + f1) * bd + f2) * bd + ol1[z1];

					f0 = (or0[s3] - or0[s2]) - (or0[s0] - or0[s1]);
					f1 = (or0[s0] - or0[s1]) - f0;
					f2 = or0[s2] - or0[s0];
					o2 = ((f0 * bd + f1) * bd + f2) * bd + or0[s1];

					f0 = (or1[z3] - or1[z2]) - (or1[z0] - or1[z1]);
					f1 = (or1[z0] - or1[z1]) - f0;
					f2 = or1[z2] - or1[z0];
					o3 = ((f0 * bd + f1) * bd + f2) * bd + or1[z1];

					*out1 = o0 + o1;
					*out2 = o2 + o3;
					break;
			}

			bpf += inc;
			if (bpf > n) bpf -= n;
			p0 = (int) bpf;
		}
		else wake = *out1 = *out2 = 0.f;
		gatebuf = *in1;

		in1++;	in2++;	in3++;	in4++;	in5++;
		out1++;	out2++;
    }
	if (!wake && !wakebuf) {
		getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_STATIC);
		getPin(PN_OUTPUT2)->TransmitStatusChange( SampleClock(), ST_STATIC);
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module::sub_process_static);
	}
    wakebuf = wake;
}






void Module::sub_process_static(long buffer_offset, long sampleFrames )
{
	sub_process(buffer_offset, sampleFrames);
	static_count = static_count - sampleFrames;
	if( static_count <= 0 ) CallHost(seaudioMasterSleepMode);
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}

void Module::open()
{
	SEModule_base::open();
	SET_PROCESS_FUNC(Module::sub_process);

	samplerate = getSampleRate();

	o = 0.f;

	//	32 due to 5 volts pitch normalised.. 2 ^ 5 = 32
	incrate = (44100.f / 32.f) / samplerate;

	samples = 0;
	actualsamples = 0;

	p = 0;
	bpf = p0 = p1 = pfr = 0;

	pf = 0.f;
	inc = tinc = 1.f;


	for (i = 1; i <= 16; i++) {
		int le = powtwo[i];	int le2 = le / 2;
		o = pi / (float)le2;
		zr[i] = cos(o);
		zi[i] = -sin(o);
	}

	n = powtwo[buffer];
	nm1 = n - 1;
	nd2 = n / 2;
	o = pi2 / (float)n;
	for (i = 0; i < n; i++) sine[i] = (1.f - cos((float)i * o)) * .5f;
	hb = n / 2;

	for (i = 0; i < 32768; i++) ol0[i] = ol1[i] = or0[i] = or1[i] = 0;

	gatebuf = 0.f;
	pitchbuf = timebuf = -77.7f;
	wakebuf = 0;

	play = 0;
	left = right = 0;

	getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_STATIC);
	getPin(PN_OUTPUT2)->TransmitStatusChange( SampleClock(), ST_STATIC);
	static_count = getBlockSize();
	SET_PROCESS_FUNC(Module::sub_process_static);
}

Module::~Module()
{
}

bool Module::getModuleProperties (SEModuleProperties* properties)
{
	properties->name = "sifft";
	properties->id = "rl sifft";
	properties->about = "sifft by rurik leffanta october 12th 2012\n\n\n";
	return true;
}

bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index ) {
	case 0:
		properties->name				= "gate";
		properties->variable_address	= &input1_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 1:
		properties->name				= "pitch";
		properties->variable_address	= &input2_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "5";
		break;
	case 2:
		properties->name				= "time";
		properties->variable_address	= &input3_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "5";
		break;
	case 3:
		properties->name				= "position";
		properties->variable_address	= &input4_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 4:
		properties->name				= "rotate";
		properties->variable_address	= &input5_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;

	case 5:
		properties->name				= "file";
		properties->variable_address	= &filename;
		properties->direction			= DR_IN;
		properties->datatype			= DT_TEXT;
		properties->default_value		= ".wav";
		properties->flags				= IO_FILENAME;
		properties->datatype_extra		= "wav";
		break;
	case 6:
		properties->name				= "buffer";
		properties->variable_address	= &buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		properties->datatype_extra		= "2=1, 4=2, 8=3, 16=4, 32=5, 64=6, 128=7, 256=8, 512=9, 1024=10, 2048=11, 4096=12, 8192=13, 16384=14, 32768=15";
		properties->default_value		= "12";
		break;
	case 7:
		properties->name				= "interpolation";
		properties->variable_address	= &interpolation;
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		properties->datatype_extra		= "floor, linear, cubic";
		break;
	case 8:
		properties->name				= "ifft interp.";
		properties->variable_address	= &ifftinterpolation;
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		properties->datatype_extra		= "floor, linear";
		break;
	case 9:
		properties->name				= "mode";
		properties->variable_address	= &mode;
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		properties->datatype_extra		= "play, scan";
		break;

	case 10:
		properties->name				= "out l";
		properties->variable_address	= &output1_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 11:
		properties->name				= "out r";
		properties->variable_address	= &output2_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;


	default:
		return false;
	};

	return true;
}

void Module::OnPlugStateChange(SEPin *pin)
{
	if	(pin->getPinID() == PN_INPUT1) {
		getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_RUN);
		getPin(PN_OUTPUT2)->TransmitStatusChange( SampleClock(), ST_RUN);
		mode ? SET_PROCESS_FUNC(Module::sub_process_scan) : SET_PROCESS_FUNC(Module::sub_process);
		wake = 1;
	}

	if	(pin->getPinID() == PN_MODE) {
		mode ? SET_PROCESS_FUNC(Module::sub_process_scan) : SET_PROCESS_FUNC(Module::sub_process);
	}

	if	(pin->getPinID() == PN_FILENAME) {
		play = 0;	samples = 0;	actualsamples = 0;
		left = right = 0;
		FILE *fp;
		fp = fopen(filename,"rb");
		if (fp) {
			char id[4];
			DWORD size;
			short wFormatTag, nBlockAlign;
			DWORD FormatChunkSize, nSamplesPerSec, nAvgBytesPerSec, DataChunkSize;
			BYTE bitl, bith, bit3;

			fread(id, sizeof(char), 4, fp);

			if (!strncmp(id, "RIFF", 4)) {
				fread(&size, sizeof(DWORD), 1, fp);
				fread(id, sizeof(char), 4, fp);
				if (!strncmp(id, "WAVE", 4)) {
					fread(id, sizeof(char), 4, fp);	//	"fmt "
					fread(&FormatChunkSize, sizeof(DWORD),1,fp);
					fread(&wFormatTag, sizeof(short), 1, fp);	//	1 = PCM
					fread(&nChannels, sizeof(short),1,fp);
					fread(&nSamplesPerSec, sizeof(DWORD), 1, fp);
					fread(&nAvgBytesPerSec, sizeof(DWORD), 1, fp);	//	SampleRate * nChannels * BitsPerSample/8
					fread(&nBlockAlign, sizeof(short), 1, fp);	//	nChannels * BitsPerSample/8
					fread(&wBitsPerSample, sizeof(short), 1, fp);	//	eg. 16, not 16 * nChannels
					fread(id, sizeof(char), 4, fp); //	"data"

					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);	//	this is an inadviseable bit of code
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);	//	that opens a few more samples
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);
					if (strncmp(id, "data", 4)) fread(id, sizeof(char), 4, fp);

					if (!strncmp(id, "data", 4)) {
						fread(&DataChunkSize, sizeof(DWORD), 1, fp); //	how many bytes of sound data
						if (DataChunkSize % 2) fread(id, sizeof(char), 1, fp);	//	padding byte
						if(nSamplesPerSec) {
							incrate = ((44100.f / 32.f) / samplerate) * (double)nSamplesPerSec / samplerate;
							pitchbuf = -77.7f;
						}

						if (wBitsPerSample == 8) {
							if (nChannels == 1) {
								samples = DataChunkSize;
								left = (float*) malloc (sizeof(float) * samples);
								right = left;
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);
									left[i] = (float)bitl / 128.f - 1.f;
							}	}
							else if (nChannels == 2) {
								samples = DataChunkSize / 2;
								left = (float*) malloc (sizeof(float) * samples);
								right = (float*) malloc (sizeof(float) * samples);
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);
									left[i] = (float)bitl / 128.f - 1.f;	right[i] = (float)bith / 128.f - 1.f;
						}	}	}
						else if (wBitsPerSample == 16) {
							if (nChannels == 1) {
								samples = DataChunkSize / 2;
								left = (float*) malloc (sizeof(float) * samples);
								right = left;
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);
									bith > 127 ? left[i] = (float)(bith * 256 + bitl - 65536) : left[i] = (float)(bith * 256 + bitl);
									left[i] /= 32768.f;
							}	}
							else if (nChannels == 2) {
								samples = DataChunkSize / 4;
								left = (float*) malloc (sizeof(float) * samples);
								right = (float*) malloc (sizeof(float) * samples);
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);
									bith > 127 ? left[i] = (float)(bith * 256 + bitl - 65536) : left[i] = (float)(bith * 256 + bitl);
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);
									bith > 127 ? right[i] = (float)(bith * 256 + bitl - 65536) : right[i] = (float)(bith * 256 + bitl);
									left[i] /= 32768.f;
									right[i] /= 32768.f;
						}	}	}
						else if (wBitsPerSample == 24) {
							if (nChannels == 1) {
								samples = DataChunkSize / 3;
								left = (float*) malloc (sizeof(float) * samples);
								right = left;
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);	fread(&bit3, sizeof(BYTE), 1, fp);
									bit3 > 127 ? left[i] = (float)(bit3 * 65536 + bith * 256 + bitl - 16777216) : left[i] = (float)(bit3 * 65536 + bith * 256 + bitl);
									left[i] /= 8388608.f;
							}	}
							else if (nChannels == 2) {
								samples = DataChunkSize / 6;
								left = (float*) malloc (sizeof(float) * samples);
								right = (float*) malloc (sizeof(float) * samples);
								for (i = 0; i < samples; i++) {
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);	fread(&bit3, sizeof(BYTE), 1, fp);
									bit3 > 127 ? left[i] = (float)(bit3 * 65536 + bith * 256 + bitl - 16777216) : left[i] = (float)(bit3 * 65536 + bith * 256 + bitl);
									fread(&bitl, sizeof(BYTE), 1, fp);	fread(&bith, sizeof(BYTE), 1, fp);	fread(&bit3, sizeof(BYTE), 1, fp);
									bit3 > 127 ? right[i] = (float)(bit3 * 65536 + bith * 256 + bitl - 16777216) : right[i] = (float)(bit3 * 65536 + bith * 256 + bitl);
									left[i] /= 8388608.f;
									right[i] /= 8388608.f;
						}	}	}
						if (samples > 0) actualsamples = samples;
					}
				}	//	else not WAVE
			}	//	else not RIFF
		}
		fclose(fp);

		if (left) {
			fftanalysis();
			play = 0;
			pf = p = 0;
		}
	}

	if	(pin->getPinID() == PN_BUFFER) {
		n = powtwo[buffer];
		invn = 1.f / (float) n;
		nm1 = n - 1;
		nd2 = n / 2;
		o = pi2 / (float)n;
		for (i = 0; i < n; i++) sine[i] = (1.f - cos((float)i * o)) * .5f;
		hb = nd2;	//	lazy use of variables, one should be removed

		if (left) {
			fftanalysis();
			play = 0;
			pf = p = 0;
		}
	}
}




void Module::dofft(float* real, float* imag) {
	int jm1, ip;
	int le, le2;
	float ur, ui;
	float tr, ti;
	j = nd2;
	for (i = 1; i <= n - 2; i++) {
		if (i < j) {
			tr = real[j];		ti = imag[j];
			real[j] = real[i];	imag[j] = imag[i];
			real[i] = tr;		imag[i] = ti;
		}
		k = nd2;
		while (k <= j) {j -= k;	k /= 2;}
		j += k;
	}
	for (k = 1; k <= buffer; k++) {
		le = powtwo[k];	le2 = le / 2;
		ur = 1.f;	ui = 0.f;
		for (j = 1; j <= le2; j++) {
			jm1	= j - 1;
			for (i = jm1; i <= nm1; i += le) {
				ip = i + le2;
				tr = real[ip] * ur - imag[ip] * ui;
				ti = real[ip] * ui + imag[ip] * ur;
				real[ip] = real[i] - tr;
				imag[ip] = imag[i] - ti;
				real[i] = real[i] + tr;
				imag[i] = imag[i] + ti;
			}
			tr = ur;
			ur = tr * zr[k] - ui * zi[k];
			ui = tr * zi[k] + ui * zr[k];
}	}	}




void Module::fftanalysis() {

	//	don't even ask me why it needs this here as well
	//	some order of pin reference in synthedit in startup makes it crash without it
	//	when the host is restarted with a previously indicated sample path
	n = powtwo[buffer];
	invn = 1.f / (float) n;
	nm1 = n - 1;
	hb = nd2 = n / 2;
	//	now back to our normally scheduled algorithm...

	if (n <= actualsamples) {
		good = 1;
		samples = actualsamples;
		//	this is not well calibrated but it seems to stop crashes at this value
		//unsigned long int cap = 536870912 / (n + 1);
		unsigned long int cap = 268435456 / (n + 1);
		if (samples > cap) samples = cap;

		int frames = (samples / hb) + 1;	//	# of frames accounts for overlap
		rl = (float*) malloc (sizeof(float) * frames * (n + 1));
		rr = (float*) malloc (sizeof(float) * frames * (n + 1));
		il = (float*) malloc (sizeof(float) * frames * (n + 1));
		ir = (float*) malloc (sizeof(float) * frames * (n + 1));

		samples /= n;
		samples = samples * n - n;

		unsigned long int fr, fa, fb, fc, frn, frhb;

		for (fr = 0; fr < frames - 1; fr++) {
			frn = fr * n;
			frhb = fr * hb;	//	first sample of frame
			for (fc = 0; fc < n; fc++) {
				fa = frn + fc;	fb = frhb + fc;
				rl[fa] = left[fb];
				rr[fa] = right[fb];
				il[fa] = ir[fa] = 0.f;
			}
		}
		frn = frames * n;
		frhb = frames * hb;
		for (fc = 0; fc < n; fc++) {
			fa = frn + fc;	fb = frhb + fc;
			if (fa < samples) {
				rl[fa] = left[fb];
				rr[fa] = right[fb];
				il[fa] = ir[fa] = 0.f;
			}
	//		else rl[fa] = rr[fa] = il[fa] = ir[fa] = 0.f;
		}

		float mx, px, rx, ix;
		for (fr = 0; fr < frames; fr++) {
			frn = fr * n;
			dofft(rl + frn, il + frn);
			dofft(rr + frn, ir + frn);
			for (i = 0; i < n; i++) {	//	cartesian to polar
				rx = rl[frn + i];
				ix = il[frn + i];
				mx = sqrt(rx * rx + ix * ix);
				if (rx == 0.f) px = 0.f;
				else {
					px = atan(ix / rx);
					if (rx < 0.f) ix < 0.f ? px -= pi : px += pi;
				}
				rl[frn + i] = mx;	il[frn + i] = px;

				rx = rr[frn + i];
				ix = ir[frn + i];
				mx = sqrt(rx * rx + ix * ix);
				if (rx == 0.f) px = 0.f;
				else {
					px = atan(ix / rx);
					if (rx < 0.f) ix < 0.f ? px -= pi : px += pi;
				}
				rr[frn + i] = mx;	ir[frn + i] = px;
			}
		}
		//	r is now magnitude, i is now phase


		//	this makes phase components the difference between current and previous
		//	except for frame 0, which is the initial phase

		for (fr = frames - 1; fr > 1; fr--) {
			frn = fr * n;
			fa = frn - n;
			for (i = 0; i < n; i++) {
				il[frn + i] = il[frn + i] - il[fa + i];
				ir[frn + i] = ir[frn + i] - ir[fa + i];
			}
		}
	}
	else good = 0;

	//	'good' added at last moment - it handles the case of the user selecting a buffer larger than the # of samples
	//	then reselecting a valid buffer length
	//	previously, if n > samples i set 'left' or 'samples' to 0, which meant the user had to reload the sample
	//	as well as set a lower buffer length.. so one more variable in the spaghetti
}